#include "DotParser.h"

#include <QFontMetrics>
#include <QStringList>
#include <QVector>

#include <cmath>

using namespace divine::gui::dot;

const QString INFO_GRAPH_KEY = QString("graph");
const QString INFO_NODE_KEY = QString("node");
const QString INFO_EDGE_KEY = QString("edge");

const QString DIRECTED_EDGE = QString("->");
const QString UNDIRECTED_EDGE = QString("--");

const int POINTS_PER_INCH = 72;

DotParser::DotParser(const QString& dot, graph::Graph& graph)
    : m_dot(dot), m_graph(graph)
{
}

void DotParser::parse()
{
    // strip the digraph { }
    int start = m_dot.indexOf('{') + 1;
    int end = m_dot.lastIndexOf('}');
    auto stripped = m_dot.mid(start, end - start);

    for (auto section : stripped.split(';'))
    {
        section = section.trimmed();

        int dataStart = section.indexOf('[');
        auto name = section.left(dataStart - 1).trimmed();
        auto data = section.mid(dataStart + 1, section.length() - dataStart - 2);

        if (name.startsWith(INFO_GRAPH_KEY))
        {
            parseGraphConfiguration(m_graphConfiguration, data);
            m_graph.setConfiguration(m_graphConfiguration);
        }
        else if (name.startsWith(INFO_NODE_KEY))
        {
            parseNodeConfiguration(m_mainNodeConfiguration, "", data);
        }
        else if (name.startsWith(INFO_EDGE_KEY))
        {
            parseEdgeConfiguration(m_mainEdgeConfiguration, "", data);
        }
        else if (name.contains(DIRECTED_EDGE) || name.contains(UNDIRECTED_EDGE))
        {
            DotEdgeConfiguration edgeConfig(m_mainEdgeConfiguration);
            parseEdgeConfiguration(edgeConfig, name, data);
            m_graph.addEdge(edgeConfig);
        }
        else
        {
            DotNodeConfiguration nodeConfig(m_mainNodeConfiguration);
            parseNodeConfiguration(nodeConfig, name, data);
            m_graph.addNode(nodeConfig);
        }
    }
}

void DotParser::parseGraphConfiguration(DotGraphConfiguration &config, const QString &data)
{
    readGraphDimensions(config, data);
}

void DotParser::parseNodeConfiguration(DotNodeConfiguration &config, const QString &name, const QString &data)
{
    readNodeBoundingRect(config, data);
    config.setBorderColor(readColor(data));
    readNodeFillColor(config, data);
    readNodeLabel(config, name, data);
}

void DotParser::parseEdgeConfiguration(DotEdgeConfiguration& config, const QString &edgeDescriptor, const QString &data)
{
    int separatorIndex = edgeDescriptor.indexOf("->");
    QString from = edgeDescriptor.left(separatorIndex - 1).trimmed();
    QString to = edgeDescriptor.mid(separatorIndex + 2).trimmed();

    readEdgeArrowheads(config, data);
    readEdgeCoordinates(config, data);
    config.setColor(readColor(data));
    readEdgeLabel(config, from, to, data);
}

void DotParser::readGraphDimensions(DotGraphConfiguration& config, const QString& data)
{
    auto dimensionData = findParameter(data, "bb");
    QVector<double> coordinates;
    if (readCoordinates(dimensionData, 4, coordinates))
    {
        config.setXOffset(coordinates[0]);
        config.setYOffset(coordinates[1]);
        double width = coordinates[2] - coordinates[0];
        double height = coordinates[3] - coordinates[1];
        config.setWidth(width);
        config.setHeight(height);
    }
}

void DotParser::readNodeBoundingRect(DotNodeConfiguration &config, const QString &data)
{
    QRectF rect = config.boundingRect();

    auto heightData = findParameter(data, "height");
    auto widthData = findParameter(data, "width");

    double height, width;
    if (readDouble(heightData, height))
    {
        rect.setHeight(height * POINTS_PER_INCH);
    }
    if (readDouble(widthData, width))
    {
        rect.setWidth(width * POINTS_PER_INCH);
    }

    auto coordinateData = findParameter(data, "pos");
    QVector<double> coordinates;
    if (readCoordinates(coordinateData, 2, coordinates))
    {
        // DOT pos is in the middle of the bounding rect, shift it for Qt
        double realX = coordinates[0] - (rect.width() / 2);
        double realY = m_graphConfiguration.height() - coordinates[1] - (rect.height() / 2);

        rect.moveLeft(realX);
        rect.moveTop(realY);
    }
    config.setBoundingRect(rect);
}

void DotParser::readNodeFillColor(DotNodeConfiguration& config, const QString &data)
{
    auto fillData = findParameter(data, "fillcolor");
    QColor backgroundColor = QColor(fillData);
    if (backgroundColor.isValid())
    {
        config.setBackgroundColor(backgroundColor.light());
    }
}

void DotParser::readNodeLabel(DotNodeConfiguration &config, const QString& nodeName, const QString &data)
{
    auto labelData = findParameter(data, "label");
    if (labelData.isEmpty())
    {
        labelData = config.label();
    }
    if (!nodeName.isEmpty())
    {
        labelData = labelData.replace("\\N", nodeName);
    }

    readLabel(config, labelData, data);
}

void DotParser::readEdgeCoordinates(DotEdgeConfiguration &config, const QString &data)
{
    auto positionData = findParameter(data, "pos");
    QPointF headArrowPoint, tailArrowPoint;
    bool headPointFound = false, tailPointFound = false;
    for(QString split : positionData.split(QRegExp("\\s+"), QString::SkipEmptyParts))
    {
        bool isHeadPoint = split.startsWith("e,");
        bool isTailPoint = split.startsWith("s,");
        if (isHeadPoint || isTailPoint)
        {
            split = split.mid(2);
        }

        QPointF point;
        QVector<double> coords;
        if (!readCoordinates(split, 2, coords))
        {
            continue;
        }

        point = QPointF(coords.at(0), m_graphConfiguration.height() - coords.at(1));
        if (isHeadPoint)
        {
            headArrowPoint = point;
            headPointFound = true;
        }
        else if (isTailPoint)
        {
            tailArrowPoint = point;
            tailPointFound = true;
        }
        else
        {
            config.appendLineCoordinate(point);
        }
    }
    if (headPointFound)
    {
        config.appendLineCoordinate(headArrowPoint);
    }
    if (tailPointFound)
    {
        config.prependLineCoordinate(tailArrowPoint);
    }
}

void DotParser::readEdgeArrowheads(DotEdgeConfiguration &config, const QString &data)
{
    auto dirData = findParameter(data, "dir").toLower();
    if (dirData == "forward")
    {
        config.setHasHeadArrow(true);
        config.setHasTailArrow(false);
    }
    else if (dirData == "back")
    {
        config.setHasHeadArrow(false);
        config.setHasTailArrow(true);
    }
    else if (dirData == "both")
    {
        config.setHasHeadArrow(true);
        config.setHasTailArrow(true);
    }
    else if (dirData == "none")
    {
        config.setHasHeadArrow(false);
        config.setHasTailArrow(false);
    }
}

void DotParser::readEdgeLabel(DotEdgeConfiguration &config, const QString &headNode, const QString &tailNode, const QString &data)
{
    auto labelData = findParameter(data, "label");
    if (labelData.isEmpty())
    {
        labelData = config.label();
    }
    if (!headNode.isEmpty() && !tailNode.isEmpty())
    {
        labelData = labelData.replace("\\E", headNode + " -> " + tailNode);
        labelData = labelData.replace("\\H", headNode);
        labelData = labelData.replace("\\T", tailNode);
    }

    readLabel(config, labelData, data);

    auto positionData = findParameter(data, "lp");
    if (!positionData.isEmpty())
    {
       QVector<double> pointVector;
       if (readCoordinates(positionData, 2, pointVector))
       {
           double middleX = pointVector.at(0);
           double middleY = m_graphConfiguration.height() - pointVector.at(1);
           // we have coordinates for the center of the label, but Qt needs rectangle from top left, calculate it
           QFontMetrics metrics(config.labelFont());
           int alignmentFlags = config.labelHorizontalAlignment() | config.labelVerticalAlignment();
           QRect textBoundingRect = metrics.boundingRect(QRect(), alignmentFlags, labelData);

           double realX = middleX - (textBoundingRect.width() / 2);
           double realY = middleY - (textBoundingRect.height() / 2);
           QRectF boundingRect(realX, realY, textBoundingRect.width(), textBoundingRect.height());
           config.setlabelBoundingRect(boundingRect);
       }
    }
}

QColor DotParser::readColor(const QString &data) const
{
    auto colorData = findParameter(data, "color");
    return QColor(colorData);
}

void DotParser::readLabel(DotLabelConfiguration &config, QString& labelData, const QString &data) const
{
    Qt::AlignmentFlag horizontalAlignment = readLabelHorizontalAlignment(labelData);
    config.setLabelHorizontalAlignment(horizontalAlignment);
    config.setLabel(labelData);

    QColor labelColor = readLabelColor(data);
    if (labelColor.isValid())
    {
        config.setLabelColor(labelColor);
    }
    QFont configFont = config.labelFont();
    readLabelFont(data, configFont);
    config.setLabelFont(configFont);
}

void DotParser::readLabelFont(const QString &data, QFont& font) const
{
    auto familyData = findParameter(data, "fontname");
    if (!familyData.isEmpty())
    {
        font.setFamily(familyData);
    }
    auto sizeData = findParameter(data, "fontsize");
    bool fontSizeIsDouble = false;
    double fontSize = sizeData.toDouble(&fontSizeIsDouble);
    if (fontSizeIsDouble)
    {
        font.setPointSize(fontSize);
    }
}

QColor DotParser::readLabelColor(const QString &data) const
{
    auto colorData = findParameter(data, "fontcolor");
    return QColor(colorData);
}

Qt::AlignmentFlag DotParser::readLabelHorizontalAlignment(QString& label) const
{
    if (label.contains("\\l"))
    {
        label = label.replace("\\l", "\n");
        return Qt::AlignLeft;
    }
    if (label.contains("\\r"))
    {
        label = label.replace("\\r", "\n");
        return Qt::AlignRight;
    }
    return Qt::AlignHCenter;
}

QString DotParser::findParameter(const QString& data, const QString& parameter) const
{
    int parameterIndex = data.indexOf(parameter + "=", 0, Qt::CaseInsensitive);
    if (parameterIndex < 0)
    {
        return QString();
    }

    int leftQuatationIndex = parameterIndex + parameter.length() + 1;
    if (data[leftQuatationIndex] != '\"')
    {
        // unquated value, take the whole thing until the next separator (or the end)
        return data.mid(leftQuatationIndex, data.indexOf(",", leftQuatationIndex + 1) - leftQuatationIndex);
    }

    int rightQuatationIndex = data.indexOf("\"", leftQuatationIndex + 1);
    if (rightQuatationIndex < 0)
    {
        return QString();
    }
    return data.mid(leftQuatationIndex + 1, rightQuatationIndex - leftQuatationIndex - 1);
}

bool DotParser::readCoordinates(const QString& data, int expectedLength, QVector<double>& result) const
{
    QStringList split = data.split(",");
    if (split.length() != expectedLength)
    {
        return false;
    }

    for (auto value : split)
    {
        bool ok = false;
        double coordinate = value.toDouble(&ok);
        if (!ok)
        {
            return false;
        }
        result.append(coordinate);
    }
    return true;
}

bool DotParser::readDouble(const QString &data, double &result) const
{
    bool ok = false;
    result = data.toDouble(&ok);
    return ok;
}
