#include "Graph.h"

using namespace divine::gui::graph;

Graph::Graph(QQuickItem *parent) : QQuickItem(parent)
{
    setFlag(QQuickItem::ItemHasContents);
    setAcceptedMouseButtons(Qt::LeftButton);
    setFiltersChildMouseEvents(true);
    setAntialiasing(true);
}

Graph::~Graph()
{
    clear();
    if (m_textComponent != nullptr)
    {
        delete m_textComponent;
    }
}

void Graph::setConfiguration(const IGraphConfiguration& config)
{
    setWidth(config.width());
    setHeight(config.height());
}

void Graph::addNode(const IGraphNodeConfiguration& nodeConfig)
{
    if (nodeConfig.boundingRect().isValid())
    {
        auto node = new GraphNode(nodeConfig);
        node->setFlag(QSGNode::OwnedByParent);
        m_nodes.append(node);
        m_nodesChanged = true;

        createTextObject(nodeConfig);
    }    
}

void Graph::addEdge(const IGraphEdgeConfiguration& edgeConfig)
{
    auto edge = new GraphEdge(edgeConfig);
    edge->setFlag(QSGNode::OwnedByParent);
    m_edges.append(edge);
    m_nodesChanged = true;

    createTextObject(edgeConfig);
}

void Graph::clear()
{
    if (m_root != nullptr)
    {
        m_root->removeAllChildNodes();
    }

    for (QQuickItem* label : m_labels)
    {
        delete label;
    }

    m_nodes.clear();
    m_edges.clear();
    m_labels.clear();
    m_nodesChanged = true;
}

void Graph::componentComplete()
{
    QQuickItem::componentComplete();
    m_textComponent = new QQmlComponent(qmlEngine(this));
    m_textComponent->setData("import QtQuick 2.10; Text {}", QUrl());
}

QSGNode* Graph::updatePaintNode(QSGNode *node, UpdatePaintNodeData * data)
{
    // until this graph is actually used we don't need to do anything
    if (width() <= 0 || height() <= 0)
    {
        return nullptr;
    }

    if (m_root == nullptr)
    {
        m_root = new QSGNode();
        m_root->markDirty(QSGNode::DirtyNodeAdded);
    }

    if (m_nodesChanged)
    {
        m_nodesChanged = false;
        m_root->removeAllChildNodes();
        m_root->markDirty(QSGNode::DirtyNodeRemoved);

        for (GraphNode* node : m_nodes)
        {
            m_root->appendChildNode(node);
        }
        for (GraphEdge* edge : m_edges)
        {
            m_root->appendChildNode(edge);
        }
        m_root->markDirty(QSGNode::DirtyNodeAdded);
    }
    return m_root;
}

void Graph::mouseReleaseEvent(QMouseEvent* event)
{
    if (onMouseEvent(event->localPos()))
    {
        event->accept();
    }
}

bool Graph::childMouseEventFilter(QQuickItem* item, QEvent* event)
{
    // node labels can cover up the whole node, so we handle clicks here
    if (event->type() == QEvent::MouseButtonRelease)
    {
        auto mouseEvent = static_cast<QMouseEvent*>(event);
        QPointF position = mapFromItem(item, mouseEvent->localPos());
        if (onMouseEvent(position))
        {
            event->accept();
        }
    }
    // labels don't need any event's -> just filter them all
    return true;
}

bool Graph::onMouseEvent(const QPointF& position)
{
    for (GraphNode* node : m_nodes)
    {
        if (!node->containsPosition(position))
        {
            continue;
        }

        nodePressed(*node);
        return true;
    }
    return false;
}

void Graph::createTextObject(const IGraphLabelConfiguration& config)
{
    if (config.label().isEmpty())
    {
        return;
    }

    QQuickItem* object = static_cast<QQuickItem*>(m_textComponent->create());
    object->setParentItem(this);

    object->setProperty("x", config.labelBoundingRect().x());
    object->setProperty("y", config.labelBoundingRect().y());
    object->setWidth(config.labelBoundingRect().width());
    object->setHeight(config.labelBoundingRect().height());

    object->setProperty("text", config.label());
    object->setProperty("color", config.labelColor());
    object->setProperty("font", config.labelFont());

    object->setProperty("horizontalAlignment", QVariant(config.labelHorizontalAlignment()));
    object->setProperty("verticalAlignment", QVariant(config.labelVerticalAlignment()));

    m_labels.append(object);
}
