#ifndef GRAPH_H
#define GRAPH_H

#include <QTextOption>
#include <QVector>
#include <QQuickItem>

#include "IGraphConfiguration.h"
#include "IGraphNodeConfiguration.h"
#include "IGraphEdgeConfiguration.h"

#include "GraphNode.h"
#include "GraphEdge.h"

namespace divine::gui::graph
{
    /*!
     * \brief The root class representing the API of the whole graphical item.
     *
     * One instance of this class represents one shown graph. This graph can at
     * any time be cleared and recreated.
     *
     * This graph uses a set of interfaces which has to be implemented and set.
     * Read their documentation to see what return values are expected.
     *
     * The graph has it's own configuration and expects sets of nodes and edges.
     * Nodes and edges are internally implemented as SGNodes and should be very quick
     * to draw even for large edges.
     *
     * As there is no SGNode support for rendering text there is one Text element
     * dynamically created for every label (QPainter solution has trouble with
     * antialiasing when scaled).
     */
    class Graph : public QQuickItem
    {
        Q_OBJECT
    public:
        explicit Graph(QQuickItem* parent = nullptr);
        virtual ~Graph() override;

        void setConfiguration(const IGraphConfiguration& config);

        void addNode(const IGraphNodeConfiguration& nodeConfig);
        void addEdge(const IGraphEdgeConfiguration& edgeConfig);

        void clear();
    protected :
        virtual void componentComplete() override;
        virtual QSGNode* updatePaintNode(QSGNode* node, UpdatePaintNodeData*) override;
        virtual void mouseReleaseEvent(QMouseEvent* event) override;
        virtual bool childMouseEventFilter(QQuickItem* item, QEvent* event) override;
    signals:
        void nodePressed(const GraphNode& node);

    private:
        bool onMouseEvent(const QPointF& position);

        void createTextObject(const IGraphLabelConfiguration& config);

        QSGNode* m_root = nullptr;

        bool m_nodesChanged = false;
        QVector<GraphNode*> m_nodes;
        QVector<GraphEdge*> m_edges;
        QVector<QQuickItem*> m_labels;

        QQmlComponent* m_textComponent = nullptr;
    };
}

#endif // GRAPH_H
