#include "GraphEdge.h"

#include <cmath>

using namespace divine::gui::graph;

const int SplineSegmentCount = 64;

GraphEdge::GraphEdge(const IGraphEdgeConfiguration& config)
    : m_lineGeometry(QSGGeometry::defaultAttributes_Point2D(), 0),
    m_headArrowGeometry(QSGGeometry::defaultAttributes_Point2D(), 3),
    m_tailArrowGeometry(QSGGeometry::defaultAttributes_Point2D(), 3)
{
    setLineCoordinates(config);
    setColor(config.lineColor());
    handleArrows(config);

    markDirty(DirtyNodeAdded);
}

void GraphEdge::handleArrows(const IGraphEdgeConfiguration &config)
{
    auto coords = config.lineCoordinates();
    auto size = coords.size();
    if (config.hasHeadArrow() && size > 1)
    {
        setArrowGeometry(m_headArrowGeometry, coords[size - 2], coords[size - 1]);
        m_headArrowNode.setMaterial(&m_material);
        m_headArrowNode.setGeometry(&m_headArrowGeometry);
        appendChildNode(&m_headArrowNode);
    }
    if (config.hasTailArrow() && size > 1)
    {
        setArrowGeometry(m_tailArrowGeometry, coords[1], coords[0]);
        m_tailArrowNode.setMaterial(&m_material);
        m_tailArrowNode.setGeometry(&m_tailArrowGeometry);
        appendChildNode(&m_tailArrowNode);
    }
}

void GraphEdge::setColor(const QColor& color)
{
    m_material.setColor(color);
    m_lineNode.markDirty(DirtyMaterial);
    m_headArrowNode.markDirty(DirtyMaterial);
    m_tailArrowNode.markDirty(DirtyMaterial);
}

void GraphEdge::setLineCoordinates(const IGraphEdgeConfiguration& config)
{
    m_lineNode.setMaterial(&m_material);
    m_lineNode.setGeometry(&m_lineGeometry);

    auto coordinates = config.lineCoordinates();

    int startingOffset = config.hasTailArrow() ? 1 : 0;
    QList<QPointF> spline;
    for (int i = startingOffset; i < coordinates.size() - 2; i += 3)
    {
        QPointF startPoint = coordinates.at(i);
        QPointF control1 = coordinates.at(i + 1);
        QPointF control2 = coordinates.at(i + 2);
        QPointF endPoint = coordinates.at(i + 3);

        for (int j = 0; j < SplineSegmentCount; j++)
        {
            qreal regular = qreal(j) / qreal(SplineSegmentCount - 1);
            qreal inverted = 1 - regular;

            QPointF point = pow(inverted, 3) * startPoint
                        + 3 * pow(inverted, 2) * regular * control1
                        + 3 * inverted * pow(regular, 2) * control2
                        + pow(regular, 3) * endPoint;

            spline.append(point);
        }
    }

    if (config.hasHeadArrow())
    {
        spline.append(coordinates.last());
    }
    if (config.hasTailArrow())
    {
        spline.prepend(coordinates.first());
    }

    int totalCount = spline.size() * 2 - 1;
    m_lineGeometry.allocate(totalCount);
    m_lineGeometry.setDrawingMode(GL_LINES);
    for (int i = 0; i < totalCount; i++)
    {
        QPointF point = spline[ceil(i / 2.0)];
        m_lineGeometry.vertexDataAsPoint2D()[i].set(point.x(), point.y());
    }

    m_lineNode.markDirty(DirtyGeometry);
    appendChildNode(&m_lineNode);
}

void GraphEdge::setArrowGeometry(QSGGeometry &geometry, const QPointF &from, const QPointF &to)
{
    auto coordinates = calculateArrowPoints(from, to);
    geometry.setDrawingMode(GL_POLYGON);
    for (int i = 0; i < coordinates.size(); i++)
    {
        QPointF point = coordinates[i];
        geometry.vertexDataAsPoint2D()[i].set(point.x(), point.y());
    }
}

QVector<QPointF> GraphEdge::calculateArrowPoints(const QPointF &pointFrom, const QPointF &pointTo) const
{
    double originalAngle = atan2(pointTo.y() - pointFrom.y(), pointTo.x() - pointFrom.x());
    double addedAngle = M_PI / 4;
    double distance = -10;
    double x1 = distance * cos(originalAngle + addedAngle) + pointTo.x();
    double y1 = distance * sin(originalAngle + addedAngle) + pointTo.y();
    double x2 = distance * cos(originalAngle - addedAngle) + pointTo.x();
    double y2 = distance * sin(originalAngle - addedAngle) + pointTo.y();

    return QVector<QPointF>({ pointTo, QPointF(x1, y1), QPointF(x2, y2) });
}
