#include "GraphNode.h"

#include <algorithm>
#include <cmath>

using namespace divine::gui::graph;

const int ELLIPSE_POINTS = 360;

GraphNode::GraphNode() :
    m_ellipseBackgroundGeometry(QSGGeometry::defaultAttributes_Point2D(), ELLIPSE_POINTS),
    m_ellipseBorderGeometry(QSGGeometry::defaultAttributes_Point2D(), ELLIPSE_POINTS)
{
    m_ellipseBackgroundNode.setMaterial(&m_ellipseBackgroundMaterial);
    m_ellipseBackgroundNode.setGeometry(&m_ellipseBackgroundGeometry);
    appendChildNode(&m_ellipseBackgroundNode);

    m_ellipseBorderMaterial.setColor(QColor(Qt::black));
    m_ellipseBorderNode.setMaterial(&m_ellipseBorderMaterial);
    m_ellipseBorderNode.setGeometry(&m_ellipseBorderGeometry);
    appendChildNode(&m_ellipseBorderNode);

    markDirty(QSGNode::DirtyNodeAdded);
}

GraphNode::GraphNode(const IGraphNodeConfiguration& config) : GraphNode()
{
    if (config.backgroundColor().isValid())
    {
        setBackgroundColor(config.backgroundColor());
    }
    if (config.borderColor().isValid())
    {
        setBorderColor(config.borderColor());
    }

    setRect(config.boundingRect());
    m_label = config.label();
}

void GraphNode::setRect(const QRectF &rect)
{
    m_rect = rect;

    m_ellipseBackgroundGeometry.setDrawingMode(GL_POLYGON);
    updateEllipseGeometry(m_ellipseBackgroundGeometry, rect);
    m_ellipseBackgroundNode.markDirty(QSGNode::DirtyGeometry);

    m_ellipseBorderGeometry.setDrawingMode(GL_LINE_LOOP);
    updateEllipseGeometry(m_ellipseBorderGeometry, rect);
    m_ellipseBorderNode.markDirty(QSGNode::DirtyGeometry);
}

const QRectF& GraphNode::getRect() const
{
    return m_rect;
}

const QString& GraphNode::getLabel() const
{
    return m_label;
}

void GraphNode::setBorderColor(const QColor &color)
{
    if (color != m_ellipseBorderMaterial.color()) {
        m_ellipseBorderMaterial.setColor(color);
        m_ellipseBorderNode.markDirty(QSGNode::DirtyMaterial);
    }
}

void GraphNode::setBackgroundColor(const QColor &color)
{
    if (color != m_ellipseBackgroundMaterial.color()) {
        m_ellipseBackgroundMaterial.setColor(color);
        m_ellipseBackgroundNode.markDirty(QSGNode::DirtyMaterial);
    }
}

bool GraphNode::containsPosition(const QPointF& position) const
{
    // fast check first
    if (!m_rect.contains(position))
    {
        return false;
    }

    double offsetX = position.x() - m_rect.x();
    double offsetY = position.y() - m_rect.y();
    // consider x set and check if y fits the ellipse
    double height = m_rect.height() / 2;
    double width = m_rect.width() / 2;
    double y1 = height * sqrt(1 - pow((offsetX - width) / width, 2)) + height;
    double y2 = m_rect.height() - y1;
    // x is set, so check if y is between - and + ellipse border
    return y2 <= offsetY && offsetY <= y1;
}

void GraphNode::updateEllipseGeometry(QSGGeometry& geometry, const QRectF &rect)
{
    if (rect.width() <= 0 || rect.height() <= 0)
    {
        return;
    }

    for (int i = 0; i < ELLIPSE_POINTS; i++)
    {
        double rad = (i - 90) * M_PI / 180.0;
        float nx = std::cos(rad) * (rect.width() / 2) + rect.x() + rect.width() / 2;
        float ny = std::sin(rad) * (rect.height() / 2) + rect.y() + rect.height() / 2;
        geometry.vertexDataAsPoint2D()[i].set(nx, ny);
    }
}
