#ifndef SETTINGSMODEL_H
#define SETTINGSMODEL_H

#include <QMap>
#include <QSettings>
#include <QString>

#include <vector>

#include "divine/dbg/info.hpp"

#include "ShortcutsModel.h"

namespace divine::gui::models
{
    /*!
     * \brief Represents the application settings.
     *
     * Currently only handles which components are being stepped into
     * and hold the settings model. Selected components are saved between sessions.
     *
     * Since this is such a small model, it was made as an QObject and
     * can be directly bound from the view without a need for a viewmodel.
     * If it grows beyond reasonable size a separate viewmodel should be extracted.
     */
    class SettingsModel : public QObject
    {
        Q_OBJECT
        Q_PROPERTY(ShortcutsModel* shortcuts READ getShortcuts CONSTANT)
        Q_PROPERTY(int componentCount READ componentCount CONSTANT)
    public:        
        explicit SettingsModel(QObject* parent = nullptr);

        ShortcutsModel* getShortcuts();

        /*!
         * \brief Gets the total number of components.
         */
        int componentCount() const;
        /*!
         * \brief Gets those components that should be debugged.
         */
        std::vector<dbg::Component> debuggedComponents();

    public slots:
        /*!
         * \brief Returns a user readable name for a component.
         * \param Index of the component.
         */
        QString getComponentName(int index) const;
        /*!
         * \brief Returns whether the component is currently set to be dubugged.
         * \param Index of the component.
         */
        bool isComponentIncluded(int index) const;
        /*!
         * \brief Set whether or not the component should be debugged.
         * \param index - index of the component
         * \param value - true if the component should be debugged
         */
        void setComponentIncluded(int index, bool value);

    private:
        void initializeComponents();
        void initializeComponent(dbg::Component component, QString name, bool defaultValue);

        QSettings m_settings;
        ShortcutsModel m_shortcuts;

        QList<dbg::Component> m_components;
        QMap<dbg::Component, QString> m_componentNames;
        QMap<dbg::Component, bool> m_includedComponents;
    };
}

#endif // SETTINGSMODEL_H
