#include "ShortcutsModel.h"

#include <QString>
#include <QMetaProperty>

using namespace divine::gui::models;

ShortcutsModel::ShortcutsModel(QSettings& settings, QObject* parent) : QAbstractItemModel(parent),
    m_settings(settings), m_categories(mapCompare), m_propertyTranslations(mapCompare), m_defaultValues(mapCompare)
{
    initializeCategoryNames();
    initializeShortcuts();
}

ShortcutsModel::~ShortcutsModel()
{
    save();
}

QHash<int, QByteArray> ShortcutsModel::roleNames() const
{
    return QHash<int, QByteArray>
    {
        { NameRole, "name" },
        { ValueRole, "value" },
        { CategoryRole, "category" }
    };
}

QModelIndex ShortcutsModel::index(int row, int column, const QModelIndex &parent) const
{
    return createIndex(row, column, nullptr);
}

QModelIndex ShortcutsModel::parent(const QModelIndex &index) const
{
    return QModelIndex();
}

int ShortcutsModel::rowCount(const QModelIndex &parent) const
{
    return parent.isValid() ? 0 : (metaObject()->propertyCount() - 1);
}

int ShortcutsModel::columnCount(const QModelIndex &parent) const
{
    return parent.isValid() ? 0 : roleNames().count();
}

QVariant ShortcutsModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
    {
        return QVariant();
    }

    QMetaProperty property = metaObject()->property(index.row() + 1);
    switch(role){
        case (NameRole):
            return m_propertyTranslations.at(property);
        case (ValueRole):
            return property.read(this).value<QKeySequence>();
        case (CategoryRole):
            return m_categoryNames.at(m_categories.at(property));
    }
    return QVariant();
}

bool ShortcutsModel::mapCompare(const QMetaProperty &lhs, const QMetaProperty &rhs)
{
    return lhs.name() < rhs.name();
}

void ShortcutsModel::initializeCategoryNames()
{
    m_categoryNames.insert(std::make_pair(General, tr("General")));
    m_categoryNames.insert(std::make_pair(Views, tr("Views")));
    m_categoryNames.insert(std::make_pair(States, tr("States")));
    m_categoryNames.insert(std::make_pair(Navigation, tr("Navigation")));
}

void ShortcutsModel::initializeShortcuts()
{
    initializeShortcut("newSession", General, tr("Start new session"), QKeySequence::New);
    initializeShortcut("addFile", General, tr("Add file"), QKeySequence::Open);
    initializeShortcut("quitApplication", General, tr("Quit application"), QKeySequence::Quit);

    initializeShortcut("switchBitcode", Views, tr("Switch between code/bitcode"), QKeySequence("Ctrl+B"));
    initializeShortcut("showDataTab", Views, tr("Show data tab"), QKeySequence("Ctrl+D"));
    initializeShortcut("showNavigationTab", Views, tr("Show navigation tab"), QKeySequence("Ctrl+F"));

    initializeShortcut("showStateGraph", States, tr("Show state graph"), QKeySequence("Ctrl+G"));
    initializeShortcut("captureState", States, tr("Capture state"), QKeySequence("Ctrl+S"));
    initializeShortcut("goToLastState", States, tr("Go to the last state"), QKeySequence("Ctrl+L"));

    initializeShortcut("callContinue", Navigation, tr("Call continue"), Qt::Key_F5);
    initializeShortcut("setBreakpoint", Navigation, tr("Set breakpoint"), Qt::Key_F9);
    initializeShortcut("stepLineOver", Navigation, tr("Step over (line)"), Qt::Key_F10);
    initializeShortcut("stepLineIn", Navigation, tr("Step into (line)"), Qt::Key_F11);
    initializeShortcut("stepLineOut", Navigation, tr("Step out (line)"), Qt::Key_F12);
    initializeShortcut("stepInstructionOver", Navigation, tr("Step over (instruction)"), QKeySequence("Ctrl+F10"));
    initializeShortcut("stepInstructionIn", Navigation, tr("Step into (instruction)"), QKeySequence("Ctrl+F11"));
    initializeShortcut("stepInstructionOut", Navigation, tr("Step out (instruction)"), QKeySequence("Ctrl+F12"));
    initializeShortcut("stepAtomicOver", Navigation, tr("Step over (atomic)"), QKeySequence("Shift+F10"));
    initializeShortcut("stepAtomicIn", Navigation, tr("Step into (atomic)"), QKeySequence("Shift+F11"));
    initializeShortcut("stepAtomicOut", Navigation, tr("Step out (atomic)"), QKeySequence("Shift+F12"));
}

void ShortcutsModel::initializeShortcut(const QString &propertyName, Category category,
    const QString &translation, const QKeySequence &defaultShortcut)
{
    int propertyIndex = metaObject()->indexOfProperty(propertyName.toLatin1().data());
    QMetaProperty property = metaObject()->property(propertyIndex);

    m_categories.insert(std::make_pair(property, category));
    m_propertyTranslations.insert(std::make_pair(property, translation));
    m_defaultValues.insert(std::make_pair(property, defaultShortcut));

    loadProperty(property, defaultShortcut);
}


void ShortcutsModel::loadProperty(QMetaProperty& property, const QKeySequence& defaultSequence)
{
    QVariant value = m_settings.value(property.name(), defaultSequence);
    if (!value.isNull())
    {
        property.write(this, value);
    }
}

void ShortcutsModel::save() const
{
    for(int i = 1; i < metaObject()->propertyCount(); i++)
    {
        QMetaProperty property = metaObject()->property(i);
        QVariant value = property.read(this);
        m_settings.setValue(property.name(), value);
    }
}

void ShortcutsModel::setShortcutValue(int shortcutIndex, QList<int> keys)
{
    QKeySequence sequence(keys.value(0, 0), keys.value(1,0), keys.value(2,0), keys.value(3,0));
    QMetaProperty property = metaObject()->property(shortcutIndex + 1);
    property.write(this, sequence);
    QModelIndex index = createIndex(shortcutIndex, 0, nullptr);
    dataChanged(index, index, QVector<int> { ValueRole });
}

void ShortcutsModel::resetShortcutValue(int shortcutIndex)
{
    QMetaProperty property = metaObject()->property(shortcutIndex + 1);
    QKeySequence sequence = m_defaultValues[property];
    property.write(this, sequence);
    QModelIndex index = createIndex(shortcutIndex, 0, nullptr);
    dataChanged(index, index, QVector<int> { ValueRole });
}
