#ifndef SHORTCUTSMODEL_H
#define SHORTCUTSMODEL_H

#include <QAbstractItemModel>
#include <QSettings>
#include <QKeySequence>

#include <map>

namespace divine::gui::models
{
    /*!
     * \brief This class represents all the shortcuts used through the application.
     *
     * This class handles the access, setting and storage of all of the shortcuts.
     * Since the functionality of the model would cover most of the viewmodels
     * responsibilities, it also implements QAbstractItemModel and allows direct
     * displaying in the settings window.
     *
     * The shortcuts are divided into categories. Shortcuts in one category often
     * share a signal, since not many slots are going to be connected to them anyway.
     */
    class ShortcutsModel : public QAbstractItemModel
    {
        Q_OBJECT
        Q_PROPERTY(QKeySequence newSession MEMBER m_newSession NOTIFY newSessionChanged)
        Q_PROPERTY(QKeySequence addFile MEMBER m_addFile NOTIFY openFileChanged)
        Q_PROPERTY(QKeySequence quitApplication MEMBER m_quitApplication NOTIFY quitApplicationChanged)

        Q_PROPERTY(QKeySequence switchBitcode MEMBER m_switchBitcode NOTIFY switchBitcodeChanged)
        Q_PROPERTY(QKeySequence showDataTab MEMBER m_showDataTab NOTIFY showTabChanged)
        Q_PROPERTY(QKeySequence showNavigationTab MEMBER m_showNavigationTab NOTIFY showTabChanged)

        Q_PROPERTY(QKeySequence showStateGraph MEMBER m_showStateGraph NOTIFY stateChanged)
        Q_PROPERTY(QKeySequence captureState MEMBER m_captureState NOTIFY stateChanged)
        Q_PROPERTY(QKeySequence goToLastState MEMBER m_goToLastState NOTIFY stateChanged)

        Q_PROPERTY(QKeySequence callContinue MEMBER m_callContinue NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence setBreakpoint MEMBER m_setBreakpoint NOTIFY setBreakpointChanged)
        Q_PROPERTY(QKeySequence stepLineOver MEMBER m_stepLineOver NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepLineIn MEMBER m_stepLineIn NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepLineOut MEMBER m_stepLineOut NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepInstructionOver MEMBER m_stepInstructionOver NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepInstructionIn MEMBER m_stepInstructionIn NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepInstructionOut MEMBER m_stepInstructionOut NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepAtomicOver MEMBER m_stepAtomicOver NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepAtomicIn MEMBER m_stepAtomicIn NOTIFY stepChanged)
        Q_PROPERTY(QKeySequence stepAtomicOut MEMBER m_stepAtomicOut NOTIFY stepChanged)
    public:
        explicit ShortcutsModel(QSettings& settings, QObject* parent = nullptr);
        ~ShortcutsModel();

        // QAbstractItemModel overloads left undocumented - the Qt documentation is clear
        QHash<int, QByteArray> roleNames() const override;
        QModelIndex index(int row, int column, const QModelIndex &parent = QModelIndex()) const override;
        QModelIndex parent(const QModelIndex &index) const override;
        int rowCount(const QModelIndex &parent = QModelIndex()) const override;
        int columnCount(const QModelIndex &parent = QModelIndex()) const override;
        QVariant data(const QModelIndex &index, int role) const override;

    public slots:
        /*!
         * \brief Sets the shortcut at the given index to the combination of the given keys.
         * \param shortcutIndex - the row the shortcut is on
         * \param keys - keys combined with their modifier which will become the comma separated shortcut
         */
        void setShortcutValue(int shortcutIndex, QList<int> keys);
        /*!
         * \brief Sets the shortcut at the given index to it's default value.
         * \param The row of the shortcut.
         */
        void resetShortcutValue(int shortcutIndex);

    signals:
        // all of the signals just notify their users of a value change
        void newSessionChanged() const;
        void openFileChanged() const;
        void quitApplicationChanged() const;
        void switchBitcodeChanged() const;
        void showTabChanged() const;
        void setBreakpointChanged() const;
        void stateChanged() const;
        void stepChanged() const;

    private:
        // different information for each shortcut
        enum Role {
            NameRole = Qt::DisplayRole + 1,
            ValueRole,
            CategoryRole
        };

        enum Category {
            General,
            Views,
            States,
            Navigation
        };

        void initializeCategoryNames();
        void initializeShortcuts();
        void initializeShortcut(const QString& propertyName, Category category,
            const QString& translation, const QKeySequence& defaultSequence);

        void loadProperty(QMetaProperty& property, const QKeySequence& defaultSequence);
        void save() const;

        QSettings& m_settings;

        // QMetaProperty doesn't include < operator, so we have to extend it
        static bool mapCompare(const QMetaProperty& lhs, const QMetaProperty& rhs);
        typedef bool(*CompareFunctor)(const QMetaProperty&, const QMetaProperty&);

        std::map<Category, QString> m_categoryNames;
        std::map<QMetaProperty, Category, CompareFunctor> m_categories;
        std::map<QMetaProperty, QString, CompareFunctor> m_propertyTranslations;
        std::map<QMetaProperty, QKeySequence, CompareFunctor> m_defaultValues;

        QKeySequence m_newSession;
        QKeySequence m_addFile;
        QKeySequence m_quitApplication;

        QKeySequence m_switchBitcode;
        QKeySequence m_showDataTab;
        QKeySequence m_showNavigationTab;

        QKeySequence m_showStateGraph;
        QKeySequence m_captureState;
        QKeySequence m_goToLastState;

        QKeySequence m_callContinue;
        QKeySequence m_setBreakpoint;
        QKeySequence m_stepLineOver;
        QKeySequence m_stepLineIn;
        QKeySequence m_stepLineOut;
        QKeySequence m_stepInstructionOver;
        QKeySequence m_stepInstructionIn;
        QKeySequence m_stepInstructionOut;
        QKeySequence m_stepAtomicOver;
        QKeySequence m_stepAtomicIn;
        QKeySequence m_stepAtomicOut;
    };
}

#endif // SHORTCUTSMODEL_H
