#include "SimulatorModel.h"

#include <divine/dbg/node.hpp>
#include <divine/dbg/stepper.hpp>
#include <divine/dbg/context.hpp>
#include <divine/dbg/print.hpp>
#include <divine/vm/eval.hpp>
#include <divine/vm/value.hpp>
#include <divine/vm/pointer.hpp>
#include <divine/mc/draw.hpp>

#include <brick-fs>

#include <QUrl>
#include <QtConcurrent/QtConcurrent>

using namespace divine::gui::models;

const std::string SimulatorModel::raw_format = "raw_value";

SimulatorModel::SimulatorModel(QObject* parent) : QObject(parent)
{
    // TODO handle context destruction
}

const QString& SimulatorModel::status() const
{
    return m_status;
}

void SimulatorModel::setStatus(const QString& status)
{
    m_status = status;
    statusChanged();
}

bool SimulatorModel::isFileLoaded() const
{
    return m_fileLoaded;
}

void SimulatorModel::setFileLoaded(bool fileLoaded)
{
    m_fileLoaded = fileLoaded;
}

bool SimulatorModel::isWorking() const
{
    return m_isWorking;
}

void SimulatorModel::setWorking(bool isWorking)
{
    m_isWorking = isWorking;
    workingChanged();
}

std::pair<QString, int> SimulatorModel::getBitcode()
{
    QString bitcode;
    int lineNumber = 0;

    vm::Eval<dbg::DNContext<vm::CowHeap>> eval(m_cli->_ctx);
    vm::CodePointer iter = m_cli->_ctx.get(_VM_CR_PC).pointer;
    vm::CodePointer origpc = iter;
    auto &prog = m_cli->_ctx.program();

    int currentLine = 0;
    for ( iter.instruction( 0 ); prog.valid( iter ); iter = iter + 1 )
    {
        currentLine++;

        auto &i = prog.instruction( iter );
        eval._instruction = &i;
        m_cli->_ctx.set( _VM_CR_PC, iter );

        if (iter == vm::CodePointer(m_cli->_ctx.get(_VM_CR_PC).pointer))
        {
            lineNumber = currentLine;
        }

        auto printer = dbg::printer(m_cli->_ctx.debug(), eval);
        if ( i.opcode == vm::lx::OpBB )
        {
            auto iop = m_cli->_ctx.debug().find( nullptr, iter + 1 ).first;
            auto stdValue = printer.value(iop->getParent());
            //auto stdValue = dbg::print::value( m_cli->_ctx.debug(), eval, iop->getParent() );
            bitcode = bitcode % QString::fromStdString(stdValue) % QString(":\n");
        }
        else
        {
            auto stdInstruction = printer.instruction(4);
            //auto stdInstruction = dbg::print::instruction( m_cli->_ctx.debug(), eval, 4 );
            bitcode = bitcode % "  " % QString::fromStdString(stdInstruction) % QString("\n");
        }
    }
    m_cli->_ctx.set(_VM_CR_PC, origpc);

    return std::pair(bitcode, lineNumber);
}

void SimulatorModel::loadFile(const QString &path)
{
    QString clearedPath = QUrl(path).path();
    _file = clearedPath.toStdString();

    QtConcurrent::run([&, clearedPath] {
        setWorking(true);
        try
        {
            bool isReport = brick::fs::takeExtension(_file) == ".report";
            if (isReport)
            {
                this->_load_report = true;
                setStatus(tr("Loading report ") + clearedPath);
            }
            else
            {
                setStatus(tr("Compiling ") + clearedPath);
            }

            statusChanged();
            setup();
            setStatus(tr("Preparing environment"));
            m_cli = std::make_unique<sim::CLI>(bitcode());
            setFileLoaded(true);

            onSettingsChanged();

            if (isReport)
            {
                loadReport();
            }
            else
            {
                start();
            }

            setShownNode(getRootNode());
            stackTraceChanged();
            statesChanged();
            onSettingsChanged();
            setStatus("");
        }
        catch (const std::exception &e)
        {
            setStatus("");
            setWorking(false);
            error(e.what(), true);
        }
        catch (...)
        {
            setStatus("");
            setWorking(false);
            error(tr("Unknown error occured"), true);
        }

        setWorking(false);
    });
}

void SimulatorModel::start()
{
    setStatus(tr("Starting target"));
    sim::command::Start command;
    m_cli->go(command);
}

void SimulatorModel::loadReport()
{
    setStatus(tr("Loading trace"));
    m_cli->trace( *_trace, false, true, [&]() { m_cli->reach_error(); } );
}

void SimulatorModel::onSettingsChanged()
{
    if (m_fileLoaded)
    {
        sim::command::Setup command;
        command.debug_components = m_settings->debuggedComponents();
        m_cli->go(command);
    }
}

void SimulatorModel::step(SimulatorModel::StepType type, SimulatorModel::StepMode mode, int count)
{
    switch (type)
    {
        case SimulatorModel::Line:
            {
                sim::command::Step step;
                runStep<sim::command::Step>(step, mode, count);
            }
            break;
        case SimulatorModel::Instruction:
            {
                sim::command::StepI stepI;
                runStep<sim::command::StepI>(stepI, mode, count);
            }
            break;
        case SimulatorModel::Atomic:
            {
                sim::command::StepA stepA;
                runStep<sim::command::StepA>(stepA, mode, count);
            }
            break;
    }
}

template <class T>
void SimulatorModel::runStep(T step, SimulatorModel::StepMode mode, int count)
{
    if (m_cli != nullptr)
    {
        step.quiet = true;
        step.count = count;
        switch (mode)
        {
            case SimulatorModel::Out:
                step.out = true;
                break;
            case SimulatorModel::Over:
                step.over = true;
                break;
        }

        int statesBefore = m_cli->_state_names.size();
        m_cli->go(step);
        setShownNode(getRootNode());
        stackTraceChanged();
        if (m_cli->_state_names.size() != statesBefore)
        {
            statesChanged();
        }
    }
}

divine::sim::DN SimulatorModel::getRootNode() const
{
    sim::DN node(m_cli->_ctx, m_cli->_ctx.snapshot());
    node.address(dbg::DNKind::Frame, m_cli->_ctx.get( _VM_CR_Frame ).pointer, true);
    return node;
}

divine::sim::DN& SimulatorModel::getShownNode() const
{
    return *m_shownNode;
}

divine::sim::DN SimulatorModel::getGlobalsNode() const
{
    return m_cli->get("$globals");
}

void SimulatorModel::setShownNode(divine::sim::DN node)
{
    m_shownNode = std::make_unique<divine::sim::DN>(node);
    shownNodeChanged();
}

QList<int> SimulatorModel::getBreakpointsForFile(const QString &file) const
{
    QList<int> lines;
    for (auto breakpointUnion : m_cli->_bps)
    {
        auto breakpointPair = breakpointUnion.get<sim::CLI::Location>();
        if (breakpointPair.first == file.toStdString())
        {
            lines.append(breakpointPair.second);
        }
    }
    return lines;
}

void SimulatorModel::addBreakpoint(const QString &file, const int line)
{
    auto location = std::make_pair(file.toStdString(), line);
    m_cli->_bps.emplace_back(location);
}

void SimulatorModel::removeBreakpoint(const QString &file, const int line)
{
    m_cli->_bps.erase(std::remove_if(std::begin(m_cli->_bps), std::end(m_cli->_bps),
                   [file, line] (sim::CLI::Breakpoint& breakpointUnion)
    {
        auto breakpointPair = breakpointUnion.get<sim::CLI::Location>();
        return breakpointPair.first == file.toStdString() && breakpointPair.second == line;
    }));
}

QList<std::pair<std::string, divine::sim::DN>> SimulatorModel::getDataNodes(sim::DN& node) const
{
    QList<std::pair<std::string, sim::DN>> retList;
    node.components([&retList](std::string name, auto sub)
    {
        retList.append(std::make_pair(name, sub));;
    });
    return retList;
}

QList<divine::sim::DN> SimulatorModel::getStackTrace() const
{
    QList<sim::DN> stackTrace;
    if (m_cli != nullptr && isFileLoaded())
    {
        sim::DN node = getRootNode();
        // max 200 levels
        for (int i = 0; i < 200; i++)
        {
            stackTrace.append(node);

            vm::value::Pointer fr(node._address);
            m_cli->_ctx.heap().skip(fr, vm::PointerBytes);
            m_cli->_ctx.heap().read(fr.cooked(), fr);
            if (fr.cooked().null())
            {
                break;
            }

            node.address(dbg::DNKind::Frame, fr.cooked());
        }
    }
    return stackTrace;
}

std::pair<llvm::StringRef, int> SimulatorModel::getFileAndLine(sim::DN node) const
{
    return m_cli->location(node.active_pc());
}

llvm::StringRef SimulatorModel::getFunctionName(sim::DN node) const
{
    return m_cli->_ctx.debug().function(node.active_pc())->getName();
}

std::map<divine::sim::Snapshot, std::string>& SimulatorModel::getStates()
{
    return m_cli->_state_names;
}

void SimulatorModel::captureState()
{
    try
    {
        m_cli->newstate(m_cli->_ctx.snapshot(), false, true);
        stackTraceChanged();
        statesChanged();
    }
    catch (const std::exception &e)
    {
        error(e.what(), false);
    }
    catch (...)
    {
        error(tr("Unknown problem occured"), false);
    }
}

void SimulatorModel::setState(const sim::Snapshot* snapshot)
{
    try
    {
        sim::command::Rewind command;
        if (snapshot != nullptr)
        {
            command.var = m_cli->_state_names[*snapshot];
        }

        m_cli->go(command);
        setShownNode(getRootNode());
        stackTraceChanged();
    }
    catch (const std::exception &e)
    {
        error(e.what(), false);
    }
}

void SimulatorModel::setLastState()
{
    setState(nullptr);
}

QFuture<QString> SimulatorModel::getDottedStateGraphAsync()
{
    return QtConcurrent::run([&] {
        setStatus(tr("Preparing graph"));
        setWorking(true);
        int statesCount = getStates().size();
        auto dot = mc::draw(bitcode(), statesCount, false);
        auto out = brick::proc::shellSpawnAndWait(brick::proc::StdinString(dot) | brick::proc::CaptureStdout, "dot" );

        auto retVal = QString::fromStdString(out.out());
        setStatus("");
        setWorking(false);
        return retVal;
    });
}
