#ifndef SIMULATORMODEL_H
#define SIMULATORMODEL_H

#include <QFuture>
#include <QObject>
#include <QString>

#include <memory>

#include "SettingsModel.h"

#include <divine/ui/cli.hpp>
#include <divine/sim/cli.hpp>
#include <divine/sim/command.hpp>

namespace divine::gui::models
{
    /*!
     * \brief Model representing the simulator's API.
     *
     * This model handles the simulation life cycle, enables overall control
     * of the simulator and is used as the center point of the whole application.
     * Apart from controlling the simulator itself it is also used to notify
     * viewmodels about which file should be displayed etc.
     * */
    class SimulatorModel : public QObject, ui::Sim
    {
        Q_OBJECT
        Q_PROPERTY(SettingsModel* settings MEMBER m_settings)
        Q_PROPERTY(bool fileLoaded READ isFileLoaded NOTIFY fileLoadedChanged)
        Q_PROPERTY(bool isWorking READ isWorking NOTIFY workingChanged)
        Q_PROPERTY(const QString& status READ status NOTIFY statusChanged)
    public:
        SimulatorModel(QObject* parent = nullptr);
        // override ui::Sim run so that we can handle things on our own
        void run() {}

        static const std::string raw_format;
        enum StepType { Line, Instruction, Atomic }; Q_ENUMS(StepType)
        enum StepMode { Over, In, Out }; Q_ENUMS(StepMode)

        /*!
         * \brief Shows whether the simulation has started.
         * <see cref="fileLoadedChanged"/> for signal notifying this property change.
         * \return false until the file was selected and is fully loaded, true afterwards
         */
        bool isFileLoaded() const;
        /*!
         * \brief Property set to true when the simulator is handling long running action.
         * <see cref="isWorkingChanged"/> for signal notifying this property change.
         */
        bool isWorking() const;
        /*!
         * \brief Property with a message about the current work being done.
         * Empty if the simulator is not working
         * <see cref="statusChanged"/> for signal notifying this property change.
         */
        const QString& status() const;

        /*!
         * \brief Gets the current LLVM bitcode.
         * \return Pair, first item is the bitcode itself, second item is the current line.
         */
        std::pair<QString, int> getBitcode();

        /*!
         * \brief Returns a copy of the current execution root.
         *
         * Root node is the node which represents the current execution state.
         */
        sim::DN getRootNode() const;
        /*!
         * \brief Returns reference of the last node that was selected by the user.
         *
         * Note that this node is not necessarily the shown file (as the shown file
         * can be overriden manually in the files view).
         */
        sim::DN& getShownNode() const;
        /*!
         * \brief Returns a node allowing access to global variables.
         */
        sim::DN getGlobalsNode() const;
        /*!
         * \brief Sets a node to be displayed. The node should be from the current call stack.
         */
        void setShownNode(sim::DN node);

        /*!
         * \brief Returns a list of line numbers on which breakpoints are placed for a file.
         * \param Path to the file.
         */
        QList<int> getBreakpointsForFile(const QString& file) const;
        void addBreakpoint(const QString& file, const int line);
        void removeBreakpoint(const QString& file, const int line);

        /*!
         * \brief Returns a list of pairs, where every first item is the name of the variable
         *  and the second item is the debug node associated with it.
         * \param Is the root node from which the variables should be gotten.
         */
        QList<std::pair<std::string, sim::DN>> getDataNodes(sim::DN& node) const;
        /*!
         * \brief Returns a list of debug nodes representing the current stack trace.
         */
        QList<sim::DN> getStackTrace() const;

        /*!
         * \brief For a given debug node returns a pair. The first item is the
         * name of the file corresponding to the given node and the second item
         * is the current line in that file.
         */
        std::pair<llvm::StringRef, int> getFileAndLine(sim::DN node) const;
        /*!
         * \brief Returns the name of the function that the given debug node
         * is currently in.
         */
        llvm::StringRef getFunctionName(sim::DN node) const;

        /*!
         * \brief Returns a map of captured program states.
         *
         * The keys are the available snapshots, whereas the values are the
         * associated names.
         */
        std::map<sim::Snapshot, std::string>& getStates();
        /*!
         * \brief Captures a new state at the given position if one hasn't been recorded yet.
         *
         * Might cause a non-fatal error, see the error signal.
         */
        void captureState();
        /*!
         * \brief Sets a current state.
         *
         * If possible the given state is set as the current execution context. That
         * involves possibly changing the current file/line (<see cref="shownNodeChanged"/>).
         * Sometimes the loading of the state might cause a non-fatal error, <see cref="error"/>).
         */
        void setState(const sim::Snapshot* snapshot);
        /*!
         * \brief Shortcut to set the last state as the current execution context.
         *
         * Also <see cref="setState"/>.
         */
        void setLastState();

        /*!
         * \brief Returns a future containing a GraphViz's dot doscription
         * of the current state graph.
         *
         * After this function is called an asynchronous operation is launched.
         * After it is finished the future will contain either a dotted graph
         * description, or an empty string if GraphViz is not installed on the current system.
         *
         * QFutureWatcher can be used to get notified when the operation finishes.
         * Simulator is set as working during this method <see cref="isWorking"/>
         */
        QFuture<QString> getDottedStateGraphAsync();
    public slots:
        /*!
         * \brief Loads the given file.
         * \param path - path to the file to be loaded.
         *
         * This method should be only called once. Once the file is loaded
         * the current application is the debugging session of the given file.
         * Anther application instance should be launched if a new file debugging
         * is needed.
         *
         * <see cref="fileLoadedChanged"/> for the signal that is emitted when
         * this function terminates. Simulator is set as working during this method
         * <see cref="isWorking"/>
         */
        void loadFile(const QString& path);
        /*!
         * \brief Performs a debugging step on the current root node and
         * notifies about the changes it brought.
         * \param type - decidedes what step method should be used
         * \param mode - allows to control where the result will be
         * \param count - if > 0 it's a number of steps to be make, otherwise is treated as continue
         *
         * Can throw the following signals:
         * <see cref="stackTraceChanged"/>
         * <see cref="shownNodeChanged"/>
         * <see cref="statesChanged"/>
         */
        void step(StepType type, StepMode mode, int count = 1);
        /*!
         * \brief Notifies this model to read the new settings from SettingsModel.
         */
        void onSettingsChanged();

    signals:
        void fileLoadedChanged();
        void statusChanged();
        void workingChanged();

        void stackTraceChanged();
        void shownNodeChanged();
        void statesChanged();

        void error(const QString& message, bool fatal);
    private:
        void setFileLoaded(bool fileLoaded);
        void setWorking(bool isWorking);
        void setStatus(const QString& status);

        void start();
        void loadReport();

        template <class T>
        void runStep(T step, StepMode mode, int count);

        std::unique_ptr<sim::CLI> m_cli;
        std::unique_ptr<sim::DN> m_shownNode;

        SettingsModel* m_settings;

        bool m_fileLoaded = false;
        bool m_isWorking = false;
        QString m_status = "";
    };
}

#endif // SIMULATORMODEL_H
