#include "CodeViewModel.h"

#include <QStringBuilder>
#include <QTextDocumentFragment>

#include <Definition>
#include <Theme>

#include "Models/SimulatorModel.h"

using namespace divine::gui::viewmodels;

CodeViewModel::CodeViewModel(QObject* parent) : QObject(parent)
{
}

divine::gui::models::SimulatorModel* CodeViewModel::simulator() const
{
    return m_simulator;
}

void CodeViewModel::setSimulator(models::SimulatorModel *simulator)
{
    m_simulator = simulator;
    connect(simulator, &models::SimulatorModel::shownNodeChanged, this, &CodeViewModel::onShownNodeChanged);
}

QQuickTextDocument* CodeViewModel::codeDocument() const
{
    return m_codeDocument;
}

void CodeViewModel::setCodeDocument(QQuickTextDocument* codeDocument)
{
    m_codeDocument = codeDocument;
    
    m_highlighter = new KSyntaxHighlighting::SyntaxHighlighter(codeDocument->textDocument());
        
    const auto theme = m_repository.defaultTheme(KSyntaxHighlighting::Repository::LightTheme);
    m_highlighter->setTheme(theme);
}

QQuickTextDocument* CodeViewModel::lineDocument() const
{
    return m_lineDocument;
}

void CodeViewModel::setLineDocument(QQuickTextDocument* lineDocument)
{
    m_lineDocument = lineDocument;
}

bool CodeViewModel::getBitcodeMode() const
{
    return m_bitcodeMode;
}

void CodeViewModel::setBitcodeMode(bool isBitcodeMode)
{
    m_bitcodeMode = isBitcodeMode;
    onShownNodeChanged();
}

int CodeViewModel::breakpointCount() const
{
    return m_breakpoints.count();
}

void CodeViewModel::displayFile(const QString& path)
{
    if (m_currentFile == path)
    {
        return;
    }

    QFile file(path);
    if (m_codeDocument != nullptr && file.exists() && file.open(QIODevice::ReadOnly | QIODevice::Text))
    {        
        QString text = QString::fromUtf8(file.readAll());
        m_codeDocument->textDocument()->setPlainText(text);        
        file.close();
    }
    else if (m_codeDocument != nullptr)
    {
        auto stdPath = path.toStdString();
        auto stringView = rt::source(stdPath);
        auto stdString = std::string(stringView);
        QString text = QString::fromStdString(stdString);
        m_codeDocument->textDocument()->setPlainText(text);
    }

    const auto definition = m_repository.definitionForFileName(path);
    m_highlighter->setDefinition(definition);
    m_highlighter->rehighlight();

    recalculateLineNumbers();

    m_breakpoints = m_simulator->getBreakpointsForFile(path);
    breakpointCountChanged(m_breakpoints.size());

    m_currentFile = path;
    // in case the file is being shown manually (from Files view)
    // decided whether or not the top stack is in that file and show
    sim::DN rootNode = m_simulator->getRootNode();
    auto rootLinePair = m_simulator->getFileAndLine(rootNode);
    if (m_currentFile == QString::fromStdString(rootLinePair.first))
    {
        m_currentLine = rootLinePair.second;
        m_isTopStack = true;
    }
    else
    {
        m_currentLine = -1;
    }
}

void CodeViewModel::displayBitcode(const QString& bitcode)
{
    if (m_codeDocument != nullptr)
    {
        m_codeDocument->textDocument()->setPlainText(bitcode);
    }
    recalculateLineNumbers();
}

void CodeViewModel::recalculateLineNumbers()
{
    auto codeDocument = m_codeDocument->textDocument();
    auto block = codeDocument->firstBlock();
    
    auto lineDocument = m_lineDocument->textDocument();
        
    QString lineNotation;
    int blockNumber = 0;
    while(block != codeDocument->lastBlock())
    {        
        if (!block.isValid())
        {
            continue;
        }
        blockNumber += 1;
        
        lineNotation = lineNotation % QString::number(blockNumber) % '\n';        
        for(int i = 1; i < block.layout()->lineCount(); i++) 
        {
            lineNotation = lineNotation % ">"  % '\n';
        }
        block = block.next();
    }    
    
    lineDocument->setPlainText(lineNotation);
}

void CodeViewModel::switchBreakpoint(int codeDocumentPosition)
{
    auto codeDocument = m_codeDocument->textDocument();
    // block number index lines 0..n while DIVINE uses 1..n, offset it
    int lineNumber = codeDocument->findBlock(codeDocumentPosition).blockNumber() + 1;
    if (!m_breakpoints.contains(lineNumber))
    {        
        m_simulator->addBreakpoint(m_currentFile, lineNumber);
        m_breakpoints.append(lineNumber);
    }
    else 
    {
        m_simulator->removeBreakpoint(m_currentFile, lineNumber);
        m_breakpoints.removeAll(lineNumber);
    }
    
    breakpointCountChanged(breakpointCount());
}

int CodeViewModel::getLineForBreakpoint(int breakpointNumber) const
{
    return m_breakpoints[breakpointNumber];
}

QRectF CodeViewModel::getRectForLine(int line) const
{
    auto codeBlock = m_codeDocument->textDocument()->findBlockByNumber(line - 1);
    auto layout = codeBlock.layout();
    if (layout != nullptr){
        auto rect = layout->boundingRect();
        auto position = layout->position();
        return QRectF(position.x(), position.y(), rect.width(), rect.height());
    }
    return QRectF();
}

int CodeViewModel::getPositionForLine(int line) const
{
    auto block = m_codeDocument->textDocument()->findBlockByLineNumber(line - 1);
    return block.isValid() ? block.position() : 0;
}

double CodeViewModel::lineHeight() const
{
    auto block = m_codeDocument->textDocument()->firstBlock();
    if(!block.isValid() || block.lineCount() < 1)
    {
        return 0;
    }
    return block.layout()->lineAt(0).height();
}

void CodeViewModel::onShownNodeChanged()
{
    if(!m_simulator->isFileLoaded())
    {
        return;
    }

    if (m_bitcodeMode)
    {
        auto bitcodePair = m_simulator->getBitcode();
        displayBitcode(bitcodePair.first);
        m_currentLine = bitcodePair.second;
        m_currentFile = "";
        m_isTopStack = true;
    }
    else
    {
        sim::DN& shownNode = m_simulator->getShownNode();

        auto shownLinePair = m_simulator->getFileAndLine(shownNode);
        QString file = QString::fromStdString(shownLinePair.first);

        displayFile(file);

        sim::DN rootNode = m_simulator->getRootNode();
        auto rootLinePair = m_simulator->getFileAndLine(rootNode);
        m_currentLine = shownLinePair.second;
        m_isTopStack = shownLinePair.first == rootLinePair.first
                && shownLinePair.second == rootLinePair.second;
    }    
    currentLineChanged();
}
