#ifndef CODEVIEWMODEL_H
#define CODEVIEWMODEL_H

#include <QObject>
#include <QList>
#include <QtQuick/QQuickTextDocument>

#include <SyntaxHighlighter>
#include <Repository>

#include "Models/SimulatorModel.h"

namespace divine::gui::viewmodels
{
    /*!
     * \brief ViewModel supporting the actual code view. Handles the displayed document,
     * switching between code and bitcode, breakpoints, syntax highlighting, etc.
     */
    class CodeViewModel : public QObject
    {
        Q_OBJECT    
        Q_PROPERTY(models::SimulatorModel* simulator READ simulator WRITE setSimulator)
        Q_PROPERTY(QQuickTextDocument* codeDocument READ codeDocument WRITE setCodeDocument)
        Q_PROPERTY(QQuickTextDocument* lineDocument READ lineDocument WRITE setLineDocument)
        Q_PROPERTY(bool bitcodeMode READ getBitcodeMode WRITE setBitcodeMode)
        Q_PROPERTY(int breakpointCount READ breakpointCount NOTIFY breakpointCountChanged)
        Q_PROPERTY(int currentLine MEMBER m_currentLine NOTIFY currentLineChanged)
        Q_PROPERTY(bool isTopStack MEMBER m_isTopStack NOTIFY currentLineChanged)
    public:
        CodeViewModel(QObject* parent = nullptr);
        
        models::SimulatorModel* simulator() const;
        void setSimulator(models::SimulatorModel* simulator);

        /*!
         * \brief Getter and setter for the code document
         * which will be used for displaying the code/bitcode.
         *
         * Syntax highlighting is done in this view model.
         */
        QQuickTextDocument* codeDocument() const;
        void setCodeDocument(QQuickTextDocument* codeDocument);
        
        /*!
         * \brief Getter and setter for the document handling the line numbering.
         *
         * For ease of use the line numbers are handled as one element and
         * are calculated in the view model.
         */
        QQuickTextDocument* lineDocument() const;
        void setLineDocument(QQuickTextDocument* codeDocument);

        /*!
         * \brief Getter and setter which shows/sets whether code or bitcode is displayed.
         */
        bool getBitcodeMode() const;
        void setBitcodeMode(bool isBitcodeMode);
        
        /*!
         * \brief Number of breakpoints in the currently displayed file.
         */
        int breakpointCount() const;
        
    public slots:                
        /*!
         * \brief Allows overriding the currently displayed path from
         *  the root node to the given file.
         * \param Path to the file.
         */
        void displayFile(const QString& path);
        /*!
         * \brief Should be called after every resize to recalculate
         * line numbers because some lines could have been wrapped.
         */
        void recalculateLineNumbers();
        
        /*!
         * \brief Switches on/off the breakpoint at the given position.
         * \param Position in the document to be used to get which line should be switched.
         */
        void switchBreakpoint(int codeDocumentPosition);
        
        /*!
         * \brief Gets line number for the breakpoint at given index.
         * \param Index of the breakpoint (0..breakpointCount()).
         */
        int getLineForBreakpoint(int breakpointNumber) const;
        /*!
         * \brief Returns bounding rectangle for the given line.
         */
        QRectF getRectForLine(int line) const;
        /*!
         * \brief Returns position in document at the beggining of the given line.
         */
        int getPositionForLine(int line) const;
        double lineHeight() const;
        
    signals:
        void breakpointCountChanged(int count);
        void currentLineChanged();

    private:        
        void onShownNodeChanged();
        void displayBitcode(const QString& bitcode);

        models::SimulatorModel* m_simulator;

        KSyntaxHighlighting::Repository m_repository;
        KSyntaxHighlighting::SyntaxHighlighter* m_highlighter;
        
        QQuickTextDocument* m_codeDocument;
        QQuickTextDocument* m_lineDocument;
        
        QList<int> m_breakpoints;

        bool m_bitcodeMode;
        bool m_isTopStack;
        QString m_currentFile;
        int m_currentLine;
    };
}
#endif // CODEVIEWMODEL_H
