#include "DataViewModel.h"

using namespace divine::gui::viewmodels;

DataViewModel::DataViewModel(QObject *parent) : QAbstractItemModel(parent)
{
}

divine::gui::models::SimulatorModel* DataViewModel::simulator() const
{
    return m_simulator;
}

void DataViewModel::setSimulator(models::SimulatorModel *simulator)
{
    m_simulator = simulator;
    QObject::connect(simulator, &models::SimulatorModel::shownNodeChanged, this, &DataViewModel::onShownDataChanged);
}

QHash<int, QByteArray> DataViewModel::roleNames() const
{
    return QHash<int, QByteArray>
    {
        { NameRole, "name" },
        { ValueRole, "value" },
        { TypeRole, "type" }
    };
}

QModelIndex DataViewModel::index(int row, int column, const QModelIndex &parent) const
{
    if (!parent.isValid())
    {
        return row < m_rootNodes.count() ? createIndex(row, column, (void*) &m_rootNodes[row]) : QModelIndex();
    }

    auto parentNodePtr = (TreeNode*) parent.internalPointer();
    initializeNode(*parentNodePtr, parent);
    return row < parentNodePtr->children.count()
            ? createIndex(row, column, (void*) &parentNodePtr->children[row]) : QModelIndex();
}

QModelIndex DataViewModel::parent(const QModelIndex &child) const
{
    auto parentNodePtr = (TreeNode*) child.internalPointer();
    return parentNodePtr->parent;
}

bool DataViewModel::hasChildren(const QModelIndex &parent) const
{
    // generally every node has at least reference to itself
    // so we can optimize this and return true for every valid index
    // -> real lazy loading
    return parent.isValid();
}

int DataViewModel::rowCount(const QModelIndex &parent) const
{
    if (!parent.isValid())
    {
        return m_rootNodes.count();
    }

    auto parentNodePtr = (TreeNode*) parent.internalPointer();
    initializeNode(*parentNodePtr, parent);
    return parentNodePtr->children.count();
}

int DataViewModel::columnCount(const QModelIndex &parent) const
{
    return 1;
}

QVariant DataViewModel::data(const QModelIndex &index, int role) const
{
    auto nodePtr = (TreeNode*) index.internalPointer();
    switch(role){
        case (NameRole):
            return QString::fromStdString(nodePtr->name);
        case (ValueRole):
            return getValueFromNode(*nodePtr);
        case (TypeRole):
            return getTypeFromNode(*nodePtr);
    }
    return QVariant();
}

void DataViewModel::onShownDataChanged()
{
    beginResetModel();

    sim::DN shownNode = m_simulator->getShownNode();
    m_rootNodes.clear();
    m_rootNodes = getChildren(shownNode);

    sim::DN globalsNode = m_simulator->getGlobalsNode();
    TreeNode globalsTreeNode(tr("globals").toStdString(), globalsNode, QModelIndex());
    m_rootNodes.append(globalsTreeNode);

    endResetModel();
    shownDataChanged();
}

QList<DataViewModel::TreeNode> DataViewModel::getChildren(sim::DN &rootNode, const QModelIndex &parent) const
{
    auto childPairs = m_simulator->getDataNodes(rootNode);
    QList<TreeNode> children = QList<TreeNode>();
    for (int i = 0; i < childPairs.count(); i++)
    {
        children.insert(i, TreeNode(childPairs[i].first, childPairs[i].second, parent));
    }
    return children;
}

void DataViewModel::initializeNode(TreeNode &node, const QModelIndex &index) const
{
    if (!node.childrenInitialized)
    {
        node.children = getChildren(node.node, index);
        node.childrenInitialized = true;
    }
}

QString DataViewModel::getValueFromNode(TreeNode &node) const
{
    QString value;
    node.node.value([&value] (std::string n, std::string stdValue)
    {
        // prefer raw values for integral types
        if (stdValue == models::SimulatorModel::raw_format || value.isEmpty())
        {
            value = QString::fromStdString(stdValue);
        }
    });
    return value;
}

QString DataViewModel::getTypeFromNode(TreeNode &node) const
{
    if (node.node._di_type != nullptr)
    {
        return QString::fromStdString(node.node.di_name(node.node._di_type));
    }
    return QString();
}
