#ifndef DATAVIEWMODEL_H
#define DATAGVIEWMODEL_H

#include <QStandardItemModel>

#include "Models/SimulatorModel.h"

namespace divine::gui::viewmodels
{
    /*!
     * \brief Represents the ViewModel for the data view.
     *
     * The data is represented as a tree using the implemented QAbstractItemModel.
     * It is a tree because every variable can of course have more variables it's
     * composed of. The children variable are expanded only when requested
     * making this a lazyly loaded tree.
     */
    class DataViewModel : public QAbstractItemModel
    {
        Q_OBJECT
        Q_PROPERTY(models::SimulatorModel* simulator READ simulator WRITE setSimulator)
    public:
        DataViewModel(QObject* parent = nullptr);

        models::SimulatorModel* simulator() const;
        void setSimulator(models::SimulatorModel* simulator);

        // overriden properties are not documented - check Qt documentation
        QHash<int, QByteArray> roleNames() const override;
        QModelIndex index(int row, int column, const QModelIndex &parent = QModelIndex()) const override;
        QModelIndex parent(const QModelIndex &child) const override;
        bool hasChildren(const QModelIndex &parent) const;
        int rowCount(const QModelIndex &parent) const override;
        int columnCount(const QModelIndex &parent) const override;
        QVariant data(const QModelIndex &index, int role) const override;
\
    signals:
        /*!
         * Additional signal notifying about data changes,
         * because the build in model reset signal doesn't trigger on the first go.
         */
        void shownDataChanged();

    private slots:
        void onShownDataChanged();

    private:
        enum Role {
            NameRole = Qt::DisplayRole + 1,
            ValueRole,
            TypeRole
        };

        struct TreeNode {
            TreeNode(std::string name, sim::DN node, QModelIndex parent)
                : name(name), node(node), parent(parent) {}

            std::string name;
            sim::DN node;
            QModelIndex parent;
            bool childrenInitialized = false;
            QList<TreeNode> children = QList<TreeNode>();
        };

        QList<TreeNode> getChildren(sim::DN& rootNode, const QModelIndex& parent = QModelIndex()) const;

        void initializeNode(TreeNode& node, const QModelIndex& index) const;
        QString getValueFromNode(TreeNode& node) const;
        QString getTypeFromNode(TreeNode& node) const;

        QList<TreeNode> m_rootNodes = QList<TreeNode>();
        models::SimulatorModel* m_simulator;
    };
}

#endif // DATAVIEWMODEL_H
