import QtQuick 2.10
import QtQuick.Controls 2.3
import QtQuick.Layouts 1.3

import divine.gui 1.0

Flickable
{
    id: rootFlickable
    anchors.fill: parent

    contentHeight: Math.max(parent.height, codeField.paintedHeight)
    flickableDirection: Flickable.VerticalFlick
    clip: true
    ScrollBar.vertical: ScrollBar {}

    RowLayout {
        id: rootLayout
        anchors.fill: parent
        spacing: 3

        Rectangle {
            id: linesRectangle
            color: 'lightGray'

            Layout.fillHeight: true
            Layout.minimumWidth: breakpointCanvas.width + lineNumberingField.width
            Layout.preferredWidth: breakpointCanvas.width + lineNumberingField.width

            RowLayout {
                anchors.fill: parent
                spacing: 0

                Canvas {
                    id: breakpointCanvas

                    width: 15
                    visible: !rootFlickable.showBitcode
                    Layout.fillHeight: true
                    Layout.preferredHeight: rootFlickable.height

                    onPaint: {
                        var context = getContext("2d");
                        context.reset();
                        context.fillStyle = Qt.rgba(1, 0, 0, 1);

                        var lineHeight = viewModel.lineHeight();
                        for(var i = 0; i < viewModel.breakpointCount; i++) {
                            var line = viewModel.getLineForBreakpoint(i);
                            var rectangle = viewModel.getRectForLine(line);
                            var ellipse = context.ellipse(0, rectangle.y, width, lineHeight);
                            ellipse.fill();
                        }
                    }
                }
                TextEdit {
                    id: lineNumberingField

                    Layout.alignment: Qt.AlignTop

                    readOnly: true
                    wrapMode: TextEdit.NoWrap
                    textFormat: TextEdit.RichText
                }
            }

            MouseArea {
                anchors.fill: parent
                enabled: !rootFlickable.showBitcode

                onPressed: {
                    if (rootFlickable.showBitcode)
                    {
                        return;
                    }

                    var position = codeField.positionAt(mouseX, mouseY);
                    viewModel.switchBreakpoint(position);
                    mouse.accepted = true;
                }
            }
        }

        Canvas {
            id: currentLineHighlighter

            Layout.fillWidth: true
            Layout.fillHeight: true

            onPaint: {
                var context = getContext("2d");
                context.reset();
                context.fillStyle = viewModel.isTopStack ?
                            Qt.rgba(0.9, 0.9, 0, 0.2) : Qt.rgba(0.4, 0.1, 0.9, 0.2);

                var rect = viewModel.getRectForLine(viewModel.currentLine);
                flickToLine(rect.y);
                context.fillRect(rect.x, rect.y, codeField.width - linesRectangle.width, rect.height);
            }

            // because TextEdit doesn't display it's cursor in readonly mode and there
            // is no way to enable it, we have to reimplement it ourselves
            Rectangle {
                id: cursorOverride

                height: codeField.cursorRectangle.height
                width: codeField.cursorRectangle.width
                x: codeField.cursorRectangle.x
                y: codeField.cursorRectangle.y
                color: 'black'
                visible: codeField.activeFocus

                Timer {
                    interval: Qt.styleHints.cursorFlashTime
                    repeat: true
                    running: codeField.activeFocus
                    onTriggered: cursorOverride.visible = !cursorOverride.visible
                    onRunningChanged: cursorOverride.visible = false
                }
            }

            TextEdit {
                id: codeField
                anchors.fill: parent

                readOnly: true
                selectByMouse: true
                selectByKeyboard: true
                persistentSelection: true
                focus: true
                wrapMode: TextEdit.Wrap                

                onWidthChanged: {
                    recalculateLineNumbers();
                    breakpointCanvas.requestPaint();
                }
            }
        }

        Action {
            enabled: !viewModel.showBitcode
            shortcut: settingsModel.shortcuts.setBreakpoint
            onTriggered: viewModel.switchBreakpoint(codeField.cursorPosition)
        }
    }

    CodeViewModel {
        id: viewModel

        codeDocument: codeField.textDocument
        lineDocument: lineNumberingField.textDocument
        bitcodeMode: rootFlickable.showBitcode

        onBreakpointCountChanged: {
            breakpointCanvas.requestPaint();
        }

        onCurrentLineChanged: {
            currentLineHighlighter.requestPaint();
            codeField.forceActiveFocus();
            codeField.cursorPosition = viewModel.getPositionForLine(viewModel.currentLine);
        }
    }

    property alias viewModel: viewModel
    property bool showBitcode: false
    property SettingsModel settingsModel

    function recalculateLineNumbers() {
        viewModel.recalculateLineNumbers();
    }

    function flickToLine(lineY)
    {
        // if the desired position is out of visible area, move to it
        var flickHeight = rootFlickable.visibleArea.heightRatio * rootFlickable.contentHeight;
        var visibleAreaTop = rootFlickable.visibleArea.yPosition * rootFlickable.contentHeight;
        // last five percent are too low
        var visibleAreaBottom = visibleAreaTop + (flickHeight * 0.95);
        if (lineY < visibleAreaTop || lineY > visibleAreaBottom)
        {
            // have ten percent leeway (unless it's on top)
            rootFlickable.contentY = lineY - (flickHeight / 10.0);
            // if the set Y position is out of bounds move back to them (top or bottom)
            rootFlickable.returnToBounds();
        }
    }
}
