import QtQuick 2.10
import QtQuick.Controls 2.3
import QtQuick.Dialogs 1.3
import QtQuick.Layouts 1.3
import Qt.labs.settings 1.0

import divine.gui 1.0

ApplicationWindow
{
    id: mainWindow
    
    visible: true
    width: 1024
    height: 760

    // We want to open the filedialog when application is fully initialised
    // if no file was given as an argument. This construction has to
    // be used because FileDialog has an issue rendering in OnCompleted.
    property bool initialised: false
    onAfterSynchronizing: {
        if (!initialised && !viewModel.initialize(simulatorModel))
        {
            openDialog.open();
        }
        initialised = true;
    }

    // Due to QTBUG-67636 which I reported actions cannot be declared
    // directly in the MenuBar so they are at the end of this file
    menuBar: MenuBar {
        id: menuBar

        height: 25
        delegate: MenuBarItem {
            height: 25
        }

        Menu {
            id: fileMenu
            title: qsTr("&File")
            enabled: !simulatorModel.isWorking

            MenuItem {
                action: newSessionAction
                ShortcutTooltip {}
            }
            MenuItem {
                action: quitAction
                ShortcutTooltip {}
            }
        }

        Menu {
            id: viewMenu
            title: qsTr("&View")

            MenuItem {
                action: showDataTabAction
                ShortcutTooltip {}
            }
            MenuItem {
                action: showNavigationTabAction
                ShortcutTooltip {}
            }
        }

        Menu {
            id: toolsMenu
            title: qsTr("&Tools")

            MenuItem { action: openOptionsAction }
        }

        Menu {
            id: helpMenu
            title: qsTr("&Help")

            MenuItem { action: openAboutAction }
        }
    }

    // actual window content
    GridLayout {
        id: mainGrid
        columns: 2
        rows: 2
        anchors.fill: parent
        enabled: !simulatorModel.isWorking

        ExecutionControlView {
            id: executionControlView
            settings: settingsModel
            viewModel.simulator: simulatorModel

            Layout.fillHeight: false
            Layout.fillWidth: true
            Layout.preferredHeight: 35

        }

        TabBar {
            id: viewsTabBar

            Layout.fillHeight: false
            Layout.fillWidth: false
            Layout.preferredHeight: 35
            Layout.alignment: Qt.AlignRight

            TabButton {
                text: qsTr("Data")
                width: implicitWidth

                ShortcutTooltip {
                    text: showDataTabAction.shortcut
                }
            }
            TabButton {
                text: qsTr("Navigation")
                width: implicitWidth

                ShortcutTooltip {
                    text: showNavigationTabAction.shortcut
                }
            }
        }

        Rectangle {
            id: codeViewBackground
            Layout.fillHeight: true
            Layout.fillWidth: true

            CodeView {
                id: codeView
                anchors.fill: parent
                visible: !stateGraph.visible
                viewModel.simulator: simulatorModel
                settingsModel: settingsModel
                showBitcode: executionControlView.showBitcode
            }
            GraphView {
                id: stateGraph
                anchors.fill: parent
                visible: false
                simulatorModel: simulatorModel
            }
        }

        Item {
            Layout.preferredWidth: parent.width * 0.3
            Layout.fillHeight: true

            StackLayout {
                anchors.fill: parent
                currentIndex: viewsTabBar.currentIndex

                ColumnLayout {
                    Loader {
                        id: dataViewLoader
                        sourceComponent: viewWrapper
                        onLoaded: {
                            item.subComponent = dataViewComponent;
                            item.title = qsTr('Data')
                        }

                        Layout.fillWidth: true
                        Layout.fillHeight: true
                    }

                    Loader {
                        id: stackTraceLoader
                        sourceComponent: viewWrapper
                        onLoaded: {
                            item.subComponent = stackTraceComponent;
                            item.title = qsTr('Stack trace')
                        }

                        Layout.fillWidth: true
                        Layout.fillHeight: true
                    }
                }
                ColumnLayout {
                    Loader {
                        id: stateViewLoader
                        sourceComponent: viewWrapper
                        onLoaded: {
                            item.subComponent = stateViewComponent;
                            item.title = qsTr('States')
                        }

                        Layout.fillWidth: true
                        Layout.fillHeight: true
                    }

                    Loader {
                        id: filesViewLoader
                        sourceComponent: viewWrapper
                        onLoaded: {
                            item.subComponent = filesViewComponent;
                            item.title = qsTr('Files')
                        }

                        Layout.fillWidth: true
                        Layout.fillHeight: true
                    }
                }
            }
        }
    }

    Rectangle {
        id: statusRectangle

        anchors.fill: parent
        color: Qt.rgba(0.75, 0.75, 0.75, 0.6)
        visible: simulatorModel.isWorking || openDialog.visible

        Label {
            anchors {
                horizontalCenter: parent.horizontalCenter
                verticalCenter: parent.verticalCenter
            }

            text: simulatorModel.status
            font.pointSize: 20
        }
    }

    // view wrapper and components
    Component {
        id: viewWrapper        

        Rectangle {
            property string title
            property Component subComponent

            border.width: 1
            border.color: 'black'

            ColumnLayout {
                anchors.fill:parent
                anchors.margins: parent.border.width

                Label {
                    text: title
                    leftPadding: 5

                    Layout.maximumHeight: 13
                }

                Loader
                {
                    id: wrappedComponentLoader
                    sourceComponent: subComponent

                    Layout.fillHeight: true
                    Layout.fillWidth: true
                }
            }
        }
    }

    Component {
        id: dataViewComponent

        DataView {
            viewModel.simulator: simulatorModel
        }
    }

    Component {
        id: stackTraceComponent
        StackTraceView {
            viewModel.simulator: simulatorModel
        }
    }

    Component {
        id: stateViewComponent
        StateView {
            settings: settingsModel
            viewModel.simulator: simulatorModel
            onStateGraphRequested: stateGraph.visible = !stateGraph.visible
        }
    }

    Component {
        id: filesViewComponent
        FilesView {
            settings: settingsModel
            viewModel.simulator: simulatorModel
            onFileRequested: codeView.viewModel.displayFile(file)
        }
    }

    // subdialogs
    FileDialog {
        id: openDialog

        selectExisting: true
        selectFolder: false
        selectMultiple: false

        nameFilters: [qsTr("C/C++ files (*.cpp *.cc *.C *.CPP *.c++ *.cp *.cxx)"),
                      qsTr("Report (*.report)"),
                      qsTr("Preprocessed (*.ii *.i *.bc *.ll *.S *.s *.o *.a)")]
        folder: shortcuts.home

        onAccepted: {
            if (!simulatorModel.fileLoaded)
            {
                simulatorModel.loadFile(openDialog.fileUrl);
            }
            else
            {
                viewModel.openNewFile(openDialog.fileUrl);
            }
        }
        // there's currently a bug which I haven't found a workaround for
        // see QTBUG-68230 for more details
        onRejected: {            
            if (!simulatorModel.fileLoaded)
            {
                Qt.quit();
            }
        }
    }

    MessageDialog {
        id: errorDialog

        property bool fatal: true
        title: qsTr("Error occured")

        onAccepted: {
            if (fatal)
            {
                Qt.quit();
            }
        }
    }

    MessageDialog {
        id: aboutDialog

        title: qsTr("About DIVINE")
        text: viewModel.getAboutText()
        informativeText: viewModel.getAdditionalText()
    }

    SettingsWindow {
        id: settingsWindow
        settingsModel: settingsModel

        onClosing: simulatorModel.onSettingsChanged()
    }

    // Models and ViewModels
    MainViewModel {
        id: viewModel
    }

    SimulatorModel {
        id: simulatorModel
        settings:  settingsModel

        onError: {
            errorDialog.text = message;
            errorDialog.fatal = fatal;
            errorDialog.open();
        }

        onShownNodeChanged: stateGraph.visible = false
    }

    SettingsModel {
        id: settingsModel
    }

    Settings {
        property alias mainWindowX: mainWindow.x
        property alias mainWindowY: mainWindow.y
        property alias mainWindowWidth: mainWindow.width
        property alias mainWindowHeight: mainWindow.height

        property alias openDialogFolder: openDialog.folder
        property alias openDialogSelectedFilter: openDialog.selectedNameFilter
    }

    // Due to QTBUG-67636 which I reported actions cannot be declared
    // directly in the MenuBar so they are declared here
    // File actions
    Action {
        id: newSessionAction
        text: qsTr("&New session...")
        onTriggered: openDialog.open()
        shortcut: settingsModel.shortcuts.newSession
    }
    Action {
        id: quitAction
        text: qsTr("&Quit")
        onTriggered: Qt.quit()
        shortcut: settingsModel.shortcuts.quitApplication
    }
    // View actions
    Action {
        id: showDataTabAction
        text: qsTr("&Data")
        shortcut: settingsModel.shortcuts.showDataTab
        onTriggered: viewsTabBar.currentIndex = 0
    }
    Action {
        id: showNavigationTabAction
        text: qsTr("&Navigation")
        shortcut: settingsModel.shortcuts.showNavigationTab
        onTriggered: viewsTabBar.currentIndex = 1
    }
    //Tools actions
    Action {
        id: openOptionsAction
        text: qsTr("O&ptions")
        onTriggered: settingsWindow.show()
    }
    // Help actions
    Action {
        id: openAboutAction
        text: qsTr("&About")
        onTriggered: aboutDialog.visible = true
    }
}
